import { omit } from '@devoxa/flocky'
import crypto from 'crypto'
import type OpenAI from 'openai'
import type { z } from 'zod'
import { zodToJsonSchema } from 'zod-to-json-schema'
import { StructuredChatError, StructuredChatErrorCode } from './errors'
import {
  ChatCompletion,
  ChatCompletionAssistantMessageParam,
  ChatCompletionCreateParams,
  ChatCompletionMessageParam,
  ChatCompletionTool,
  ChatCompletionToolChoiceOption,
  RequestOptions,
  WithOptional,
  WithRequired,
  CacheAdapter,
  OpenAIChatCompletion,
} from './types'

export * from './cacheAdapters/fileSystem'
export * from './errors'
export * from './types'

export interface StructuredChatOptions<TTools extends Array<ChatCompletionTool>> {
  openai: OpenAI
  cacheAdapter?: CacheAdapter
  params: WithOptional<ChatCompletionCreateParams<TTools>, 'messages'>
  options?: Omit<RequestOptions, '_attempt'>
}

export class StructuredChat<TTools extends Array<ChatCompletionTool>> {
  openai: OpenAI
  cacheAdapter?: CacheAdapter
  params: WithOptional<ChatCompletionCreateParams<TTools>, 'messages'>
  options?: Omit<RequestOptions, '_attempt'>

  _messages: ChatCompletionCreateParams<TTools>['messages'] = []

  constructor(options: StructuredChatOptions<TTools>) {
    this.openai = options.openai
    this.cacheAdapter = options.cacheAdapter
    this.params = options.params
    this.options = options.options
  }

  async send(
    params: WithOptional<Omit<ChatCompletionCreateParams<TTools>, 'tools'>, 'model'>,
    options?: RequestOptions
  ): Promise<ChatCompletion<TTools>> {
    const result = await this._send(params, options)

    // Add the parsed messages to the existing conversation
    // Note that we DON'T include error correction messages into the conversation
    // to prevent the model from referencing incorrectly generated messages.
    this._messages.push(...params.messages)
    this._messages.push(result.choices[0].message)

    return result
  }

  async _send(
    params: WithOptional<Omit<ChatCompletionCreateParams<TTools>, 'tools'>, 'model'>,
    options?: RequestOptions
  ): Promise<ChatCompletion<TTools>> {
    const _params = { ...this.params, ...params }
    const _options = { maxErrorCorrectionTries: 2, ...this.options, _attempt: 0, ...options }
    const tools = _params.tools || []

    // Add the messages to the existing conversation
    const messages = this._messages.concat(_params.messages)

    // Encode parameters for the OpenAI API
    // - Messages must use JSON (as strings) instead of JS data types for tool calls
    // - Tools must use JSON schemas (as strings) instead of Zod schemas
    // - Tool choice "function" must be "auto" or a specific function name
    const encodedMessages = messages.map((message) =>
      this._encodeMessage(message as ChatCompletionMessageParam<Array<ChatCompletionTool>>)
    )

    const encodedTools = _params.tools
      ? _params.tools.map((tool) => this._encodeTool(tool))
      : undefined

    const encodedToolChoice = _params.tool_choice
      ? this._encodeToolChoice(tools, _params.tool_choice)
      : undefined

    // Send the request to the OpenAI API
    // This internally already handles retrying for server errors, rate limits and timeouts
    // and converts error statuses into thrown errors, so we don't need to do it here.
    const result = await this._create(
      {
        ..._params,
        messages: encodedMessages,
        tools: encodedTools,
        tool_choice: encodedToolChoice,
      },
      omit(_options, ['maxErrorCorrectionTries', '_attempt'])
    )

    // Parse the result from the OpenAI API
    // - Messages must use JS data types instead of JSON, and they must match the function schema
    result.choices = result.choices.slice(0, 1)
    const toolCalls = result.choices[0].message.tool_calls

    const hasFunctionCall = toolCalls && toolCalls.length > 0
    const hasToHaveFunctionCall =
      _params.tool_choice === 'function' || typeof encodedToolChoice === 'object'

    if (!hasFunctionCall && hasToHaveFunctionCall) {
      const name = typeof encodedToolChoice === 'object' ? encodedToolChoice.function.name : null

      const _errorMessage = name
        ? `You did not call a function. Respond by calling the "${name}" function.`
        : `You did not call a function. Respond by calling an available function.`

      if (_options._attempt < _options.maxErrorCorrectionTries) {
        const responseMessages = [{ role: 'user' as const, content: _errorMessage }]
        return this._sendErrorCorrection(result, responseMessages, _params, _options)
      } else {
        throw new StructuredChatError('FUNCTION_CALL_NO_FUNCTION_CALLED', _errorMessage)
      }
    }

    if (hasFunctionCall) {
      const parsedToolCalls = toolCalls.map((toolCall) =>
        this._parseToolCall(tools, encodedToolChoice || 'auto', toolCall)
      )

      // Handle parse failures
      const parseFailure = parsedToolCalls.find((x) => !x.success)
      if (parseFailure && !parseFailure.success) {
        if (_options._attempt < _options.maxErrorCorrectionTries) {
          const responseMessages = parsedToolCalls.map((toolCall, index) => ({
            role: 'tool' as const,
            tool_call_id: toolCalls[index].id,
            content: !toolCall.success ? toolCall.errorMessage : 'OK',
          }))

          return this._sendErrorCorrection(result, responseMessages, _params, _options)
        } else {
          throw new StructuredChatError(parseFailure.errorCode, parseFailure.errorMessage)
        }
      }

      // Overwrite the tool calls with the parsed arguments
      for (let i = 0; i !== toolCalls.length; i++) {
        const parsedToolCall = parsedToolCalls[i]

        /* c8 ignore next 1 */
        if (!parsedToolCall.success) continue

        toolCalls[i] = {
          ...toolCalls[i],
          function: {
            ...toolCalls[i].function,
            // Note that this is not a string, it's the parsed arguments. We're fixing the types below.
            arguments: parsedToolCall.arguments as string,
          },
        }
      }
    }

    // Type the result - it's ensured to be parsed because of the above codeblock
    const parsedResult = result as ChatCompletion<TTools>
    return { ...parsedResult, usage: { ...parsedResult.usage, requests: 1 } }
  }

  async _sendErrorCorrection(
    result: OpenAI.ChatCompletion,
    responseMessages: Array<ChatCompletionMessageParam<TTools>>,
    params: WithOptional<Omit<ChatCompletionCreateParams<TTools>, 'tools'>, 'model'>,
    options: WithRequired<RequestOptions, '_attempt'>
  ) {
    const messages = [
      ...params.messages,
      result.choices[0].message as ChatCompletionAssistantMessageParam<TTools>,
      ...responseMessages,
    ]

    const correctResult = await this._send(
      { ...params, messages },
      { ...options, _attempt: options._attempt + 1 }
    )

    return {
      ...correctResult,
      usage: {
        completion_tokens: correctResult.usage.completion_tokens + result.usage!.completion_tokens,
        prompt_tokens: correctResult.usage.prompt_tokens + result.usage!.prompt_tokens,
        total_tokens: correctResult.usage.total_tokens + result.usage!.total_tokens,
        requests: correctResult.usage.requests + 1,
      },
    }
  }

  async _create(
    body: OpenAI.ChatCompletionCreateParamsNonStreaming,
    options?: OpenAI.RequestOptions
  ): Promise<OpenAIChatCompletion> {
    const hash = await this._md5({ body, options })

    if (this.cacheAdapter) {
      const value = await this.cacheAdapter.get(hash)
      if (value) return value
    }

    const result = await this.openai.chat.completions.create(body, options)

    if (this.cacheAdapter) {
      await this.cacheAdapter.set(hash, JSON.parse(JSON.stringify(result)))
    }

    return result
  }

  async _md5(value: unknown) {
    return crypto.createHash('md5').update(JSON.stringify(value)).digest('hex')
  }

  setMessages(messages: ChatCompletionCreateParams<TTools>['messages']) {
    this._messages = messages
  }

  addMessage(message: ChatCompletionCreateParams<TTools>['messages'][number]) {
    this._messages.push(message)
  }

  getMessages() {
    return this._messages
  }

  // Encode a message in our format (using structured data) or OpenAI format as OpenAI format (using JSON strings)
  _encodeMessage(
    message:
      | ChatCompletionMessageParam<Array<ChatCompletionTool>>
      | OpenAI.ChatCompletionMessageParam
  ): OpenAI.ChatCompletionMessageParam {
    if (message.role !== 'assistant') {
      return message
    }

    return {
      ...message,
      tool_calls: message.tool_calls
        ? message.tool_calls.map((toolCall) => ({
            ...toolCall,
            function: {
              ...toolCall.function,
              arguments:
                typeof toolCall.function.arguments === 'string'
                  ? toolCall.function.arguments
                  : JSON.stringify(toolCall.function.arguments),
            },
          }))
        : undefined,
    }
  }

  // Encode a tool in our format (using a Zod schema) as OpenAI format (using a JSON schema)
  _encodeTool(tool: ChatCompletionTool): OpenAI.ChatCompletionTool {
    return {
      ...tool,
      function: {
        ...tool.function,
        parameters: tool.function.parameters
          ? this._zodToJsonSchema(tool.function.parameters)
          : undefined,
      },
    }
  }

  _zodToJsonSchema(schema: z.ZodSchema) {
    const jsonSchema = zodToJsonSchema(schema)

    // Remove keys that are not required to produce a valid JSON schema to save some tokens
    return this._deepRemoveKeys(jsonSchema, ['$schema', 'additionalProperties'])
  }

  _deepRemoveKeys<T>(schema: T, exclude: Array<string>): T {
    if (schema instanceof Array) {
      return schema.map((value) => this._deepRemoveKeys(value, exclude)) as T
    }

    if (typeof schema === 'object') {
      const entries = Object.entries(schema as object)
        .filter(([key]) => !exclude.includes(key))
        .map(([key, value]) => [key, this._deepRemoveKeys(value, exclude)])

      return Object.fromEntries(entries) as T
    }

    return schema
  }

  _encodeToolChoice(
    tools: Array<ChatCompletionTool>,
    toolChoice: ChatCompletionToolChoiceOption<TTools>
  ): OpenAI.ChatCompletionToolChoiceOption {
    if (toolChoice === 'function') {
      return tools.length > 1
        ? 'auto'
        : { type: 'function', function: { name: tools[0].function.name } }
    }

    return toolChoice
  }

  _parseToolCall(
    tools: Array<ChatCompletionTool>,
    toolChoice: OpenAI.ChatCompletionToolChoiceOption,
    toolCall: OpenAI.ChatCompletionMessageToolCall
  ):
    | { success: false; errorCode: StructuredChatErrorCode; errorMessage: string }
    | { success: true; arguments: unknown } {
    const name = toolCall.function.name

    // Check if the tool referenced in the tool call is matching the tool choice
    if (typeof toolChoice === 'object' && toolChoice.function.name !== name) {
      return {
        success: false,
        errorCode: 'FUNCTION_CALL_INCORRECT_FUNCTION_CALLED',
        errorMessage: `You did not call the correct function. Call the "${toolChoice.function.name}" function instead.`,
      }
    }

    // Find the tool referenced in the tool call
    const tool = tools.find((x) => x.function.name === name)
    if (!tool) {
      return {
        success: false,
        errorCode: 'FUNCTION_CALL_NONEXISTENT_FUNCTION_CALLED',
        errorMessage: `The function "${name}" does not exist. Only use the available functions.`,
      }
    }

    // Find the schema for the tool - if there is none the function does not take
    // any arguments so we can just assume the model called it with none.
    const schema = tool.function.parameters
    if (!schema) {
      return { success: true, arguments: undefined }
    }

    // Parse the arguments as JSON
    const args = this._parseGptJson(toolCall.function.arguments)
    if (!args.success) {
      return {
        success: false,
        errorCode: 'FUNCTION_ARGUMENTS_MALFORMED_JSON',
        errorMessage:
          `The function "${name}" was called with a malformed JSON value as arguments. ` +
          `Correct the JSON value to make it parsable and call the "${name}" function again.`,
      }
    }

    // Ensure the arguments are matching the schema of the tool
    const parsedArgs = schema.safeParse(args.data)
    if (!parsedArgs.success) {
      const errors = parsedArgs.error.errors.map(
        (error) => `- \`.${error.path.join('.')}\`: ${error.message}`
      )

      return {
        success: false,
        errorCode: 'FUNCTION_ARGUMENTS_INVALID_SCHEMA',
        errorMessage:
          `The function "${name}" was called with invalid arguments. ` +
          `Apply these fixes and call the "${name}" function again:\n` +
          errors.join('\n'),
      }
    }

    return {
      success: true,
      arguments: parsedArgs.data,
    }
  }

  _parseGptJson(value: string) {
    try {
      return { success: true as const, data: JSON.parse(value) }
    } catch (_) {
      return { success: false as const }
    }
  }
}
