import type OpenAI from 'openai'
import type { z } from 'zod'

export interface ChatCompletionCreateParams<TTools extends Array<ChatCompletionTool>>
  extends Omit<
    OpenAI.Chat.Completions.ChatCompletionCreateParamsNonStreaming,
    'model' | 'messages' | 'tools' | 'tool_choice' | 'functions' | 'function_call'
  > {
  /**
   * ID of the model to use. See
   * [Function calling](https://platform.openai.com/docs/guides/function-calling/introduction)
   * for details on which models have been trained for this capability.
   */
  model: 'gpt-3.5-turbo' | 'gpt-4-turbo-preview'

  /** A list of messages to add to the conversation. */
  messages: Array<ChatCompletionMessageParam<TTools>>

  /**
   * A list of tools the model may call. Currently, only functions are supported as a
   * tool. Use this to provide a list of functions the model may generate inputs for.
   */
  tools?: TTools

  /**
   * Controls which (if any) function is called by the model.
   * - `none` means the model will not call a function and instead generate a message.
   * - `auto` means the model can pick between generating a message or calling a function.
   * - `function` means the model will always call a function and never generate a message.
   * - `{ type: 'function', function: { name: 'my_function' } }` forces the model to call a specific function.
   *
   * @default `none` when no functions are present, `auto` when functions are present.
   */
  tool_choice?: ChatCompletionToolChoiceOption<TTools>
}

export type ChatCompletionMessageParam<TTools extends Array<ChatCompletionTool>> =
  | OpenAI.ChatCompletionSystemMessageParam
  | OpenAI.ChatCompletionUserMessageParam
  | ChatCompletionAssistantMessageParam<TTools>
  | OpenAI.ChatCompletionToolMessageParam
  | OpenAI.ChatCompletionFunctionMessageParam

export interface ChatCompletionAssistantMessageParam<TTools extends Array<ChatCompletionTool>> {
  /** The role of the messages author, in this case `assistant`. */
  role: 'assistant'

  /** The contents of the assistant message. Required unless `tool_calls` is specified. */
  content?: string | null

  /**
   * An optional name for the participant. Provides the model information to
   * differentiate between participants of the same role.
   */
  name?: string

  /** The tool calls generated by the model, such as function calls. */
  tool_calls?: Array<ChatCompletionMessageToolCall<TTools>>
}

interface ChatCompletionMessageToolCall<TTools extends Array<ChatCompletionTool>> {
  /** The ID of the tool call. */
  id: string

  /** The type of the tool. Currently, only `function` is supported. */
  type: 'function'

  /** The function that the model called. */
  function: ChatCompletionMessageToolCallFunction<TTools>
}

type ChatCompletionMessageToolCallFunction<TTools extends Array<ChatCompletionTool>> =
  ChatCompletionMessageToolCallFunctionGenerator<TTools[number]>[TTools[number]['function']['name']]

type ChatCompletionMessageToolCallFunctionGenerator<
  T extends { function: { name: string; parameters: z.ZodSchema | null } },
> = {
  [K in T as K['function']['name']]: {
    /** The name of the function to call. */
    name: K['function']['name']

    /**
     * The arguments to call the function with, as generated by the model.
     * Guaranteed to be matching the provided Zod schema of the function.
     */
    arguments: K['function']['parameters'] extends null
      ? undefined
      : z.infer<NonNullable<K['function']['parameters']>>
  }
}

export interface ChatCompletionTool {
  /** The type of the tool. Currently, only `function` is supported. */
  type: 'function'

  function: ChatCompletionToolFunctionDefinition
}

interface ChatCompletionToolFunctionDefinition {
  /**
   * The name of the function to be called. Must be a-z, A-Z, 0-9, or contain
   * underscores and dashes, with a maximum length of 64.
   */
  name: string

  /**
   * A description of what the function does, used by the model to choose when and
   * how to call the function.
   */
  description: string

  /**
   * The parameters the functions accepts, described as a Zod schema. See the
   * [guide](https://platform.openai.com/docs/guides/text-generation/function-calling)
   * for examples.
   *
   * Omitting `parameters` defines a function with an empty parameter list.
   */
  parameters: z.ZodSchema | null
}

export type ChatCompletionToolChoiceOption<TTools extends Array<ChatCompletionTool>> =
  | 'none'
  | 'auto'
  | 'function'
  | ChatCompletionNamedToolChoice<TTools>

interface ChatCompletionNamedToolChoice<TTools extends Array<ChatCompletionTool>> {
  /** The type of the tool. Currently, only `function` is supported. */
  type: 'function'

  function: ChatCompletionNamedToolChoiceFunction<TTools>
}

interface ChatCompletionNamedToolChoiceFunction<TTools extends Array<ChatCompletionTool>> {
  /** The name of the function to call. */
  name: TTools[number]['function']['name']
}

export interface RequestOptions extends OpenAI.RequestOptions {
  /**
   * The maximum amount of time (in milliseconds) that the client should wait for a response
   * from the server before timing out a single request.
   *
   * Note that request timeouts are retried by default, so in a worst-case scenario you may wait
   * much longer than this timeout before the promise succeeds or fails.
   *
   * @default 60_0000
   */
  timeout?: number

  /**
   * The maximum number of times that the client will retry a request in case of a
   * temporary failure, like a network error or a 5XX error from the server.
   *
   * @default 2
   */
  maxRetries?: number

  /**
   * The maximum number of times that the client will attempt to correct errors in case of a
   * hallucinated model response, like an incorrect function name or an argument schema mismatch.
   *
   * @default 2
   */
  maxErrorCorrectionTries?: number

  // Internal options
  _attempt?: number
}

export type ChatCompletion<TTools extends Array<ChatCompletionTool>> = Omit<
  OpenAI.ChatCompletion,
  'choices' | 'usage'
> & {
  /** A list of chat completion choices. Always contains one element. */
  choices: Array<
    Omit<OpenAI.ChatCompletion.Choice, 'message'> & {
      /** A chat completion message generated by the model. */
      message: ChatCompletionAssistantMessageParam<TTools>
    }
  >

  /** Usage statistics for the completion request. */
  usage: OpenAI.ChatCompletion['usage'] & { requests: number }
}

export type OpenAIChatCompletion = OpenAI.ChatCompletion

export abstract class CacheAdapter {
  /** Get the chat completion result by the hash of the request. */
  get: (hash: string) => Promise<OpenAIChatCompletion | null>

  /** Set the chat completion result for the hash of the request. */
  set: (hash: string, value: OpenAIChatCompletion) => Promise<void>
}

// --- Helper Types ---

/** Make some keys of T optional. */
export type WithOptional<T, K extends keyof T> = Omit<T, K> & Partial<Pick<T, K>>

/** Make some keys of T required. */
export type WithRequired<T, K extends keyof T> = T & { [P in K]-?: T[P] }
