import OpenAI from 'openai'
import { StructuredChat } from '../src/index'
import { z } from 'zod'
import { CacheAdapter } from '../src/types'

const mockChatCompletionsCreate = jest.fn()
const mockOpenAi = {
  chat: { completions: { create: mockChatCompletionsCreate } },
} as unknown as OpenAI

const mockCacheAdapterGet = jest.fn()
const mockCacheAdapterSet = jest.fn()
const mockCacheAdapter = {
  get: mockCacheAdapterGet,
  set: mockCacheAdapterSet,
} as unknown as CacheAdapter

function generateMockChatCompletionCreateResult(message: OpenAI.ChatCompletionMessage) {
  return {
    id: 'chatcmpl-8m8lGcBrChV9FrLC0h7MQ3Wk1iAHf',
    model: 'gpt-4-0125-preview',
    object: 'chat.completion',
    created: 1706484114,
    system_fingerprint: 'fp_376b7f78b9',
    choices: [
      {
        finish_reason: 'stop',
        index: 0,
        logprobs: null,
        message,
      },
    ],
    usage: {
      completion_tokens: 6,
      prompt_tokens: 97,
      total_tokens: 103,
    },
  }
}

describe('StructuredChat', () => {
  beforeEach(() => {
    mockChatCompletionsCreate.mockReset()
    mockCacheAdapterGet.mockReset()
    mockCacheAdapterSet.mockReset()
  })

  test('can create a new instance', () => {
    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: { model: 'gpt-4-turbo-preview' },
    })

    expect(typeof chat.send).toEqual('function')
  })

  test('can send a natural language chat', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: "She didn't go to the market.",
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: { model: 'gpt-4-turbo-preview' },
      options: { timeout: 5 * 1000, maxRetries: 2 },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with statements, and your task is to convert them to standard English.',
        },
        { role: 'user', content: 'She no went to the market.' },
      ],
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()
  })

  test('can send a natural language chat with a message history', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'I am watching an episode of a TV show.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      cacheAdapter: mockCacheAdapter,
      params: { model: 'gpt-4-turbo-preview' },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with statements, and your task is to convert them to standard English.',
        },
        { role: 'user', content: 'She no went to the market.' },
        { role: 'assistant', content: "She didn't go to the market." },
        { role: 'user', content: 'I am watching an episode tv show.' },
      ],
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()

    expect(mockCacheAdapterGet.mock.calls).toMatchSnapshot()
    expect(mockCacheAdapterSet.mock.calls).toMatchSnapshot()
  })

  test('can manipulate the message history', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'I am watching an episode of a TV show.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: { model: 'gpt-4-turbo-preview' },
    })

    chat.setMessages([
      {
        role: 'system',
        content:
          'You will be provided with statements, and your task is to convert them to standard English.',
      },
      { role: 'user', content: 'She no went to the market.' },
    ])

    chat.addMessage({ role: 'assistant', content: "She didn't go to the market." })

    const result = await chat.send({
      messages: [{ role: 'user', content: 'I am watching an episode tv show.' }],
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
  })

  test('can send a structured chat', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { timeout: 5 * 1000, maxRetries: 2 },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()

    // The function names and arguments are strongly typed and work for correct values
    result.choices[0].message.tool_calls?.[0].function.name === 'return_sentiment' &&
      result.choices[0].message.tool_calls[0].function.arguments.sentiment

    result.choices[0].message.tool_calls?.[0].function.name === 'return_sentiment' &&
      // @ts-expect-error The function argument does not exist
      result.choices[0].message.tool_calls[0].function.arguments.foo

    // @ts-expect-error The function name does not exist
    result.choices[0].message.tool_calls?.[0].function.name === 'foo_function'

    // @ts-expect-error The deprecated function_call reference does not exist
    result.choices[0].message.function_call

    // @ts-expect-error The function arguments are not string typed
    result.choices[0].message.tool_calls?.[0].function.arguments.charAt
  })

  test('can send a structured chat with a message history', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_mVbd6mqld7DVLucdor4dQIpw',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment": "negative"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      cacheAdapter: mockCacheAdapter,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
        {
          role: 'assistant',
          content: null,
          tool_calls: [
            {
              function: {
                arguments: {
                  sentiment: 'positive',
                },
                name: 'return_sentiment',
              },
              id: 'call_2oKm3MB182UOdulkNKR5kpak',
              type: 'function',
            },
          ],
        },
        { role: 'tool', tool_call_id: 'call_2oKm3MB182UOdulkNKR5kpak', content: 'OK' },
        { role: 'user', content: 'I hated the new Superman movie!' },
      ],

      tool_choice: 'function',
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()

    expect(mockCacheAdapterGet.mock.calls).toMatchSnapshot()
    expect(mockCacheAdapterSet.mock.calls).toMatchSnapshot()
  })

  test('can send a structured chat with multiple functions', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_mVbd6mqld7DVLucdor4dQIpw',
            function: {
              name: 'allow_message',
              arguments: '',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'allow_message' as const,
              description: 'Allow this message to get past the spam filter',
              parameters: null,
            },
          },
          {
            type: 'function',
            function: {
              name: 'deny_message' as const,
              description: 'Deny the message to get past the spam filter',
              parameters: null,
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a message, and your task is to decide if that message is spam.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'function',
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()
  })

  test('can send a structured chat with multiple function calls', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_0oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
          {
            type: 'function',
            id: 'call_1oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
  })

  test('can send a semi-structured chat (not structured)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'The sentiment is positive.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'auto',
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
  })

  test('can send a semi-structured chat (structured)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'auto',
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
  })

  test('can send a structured chat (function with no arguments)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: null,
            },
          },
        ],
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()

    // The function names and arguments are strongly typed and work for correct values
    const func = result.choices[0].message.tool_calls?.[0].function
    if (func?.name === 'return_sentiment') {
      type AssertUndefined<T> = [T] extends [undefined] ? T : never
      const shouldBeUndefined: AssertUndefined<typeof func.arguments> = undefined
      expect(shouldBeUndefined).toEqual(undefined)
    }
  })

  test('can send a structured chat (no function called)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'This tweet has positive sentiment.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'function',
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can send a structured chat with multiple functions (no function called)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'This message is not spam.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'allow_message' as const,
              description: 'Allow this message to get past the spam filter',
              parameters: null,
            },
          },
          {
            type: 'function',
            function: {
              name: 'deny_message' as const,
              description: 'Deny the message to get past the spam filter',
              parameters: null,
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a message, and your task is to decide if that message is spam.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'function',
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can send a structured chat (incorrect function called)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_the_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can send a structured chat (non-existing function called)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'sentiment_return',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can send a structured chat (function arguments not matching schema)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"excited"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can send a structured chat (malformed json)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive""}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { maxErrorCorrectionTries: 0 },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
  })

  test('can error correct a structured chat (function arguments not matching schema)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_0oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"tweetSentiment":"happy"}', // Invalid
            },
          },
        ],
      })
    )

    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_1oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"happy"}', // Also invalid
            },
          },
          {
            type: 'function',
            id: 'call_3oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: {
        timeout: 5 * 1000,
        maxRetries: 2,
        maxErrorCorrectionTries: 2,
      },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()
  })

  test('can error correct a structured chat (no function called)', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'This tweet has positive sentiment.',
      })
    )

    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const result = await chat.send(
      {
        messages: [
          {
            role: 'system',
            content:
              'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
          },
          { role: 'user', content: 'I loved the new Batman movie!' },
        ],

        tool_choice: 'function',
      },
      { maxErrorCorrectionTries: 2 }
    )

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toMatchSnapshot()
  })

  test('can run out of error correction tries', async () => {
    mockChatCompletionsCreate.mockResolvedValue(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'This message is not spam.',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'allow_message' as const,
              description: 'Allow this message to get past the spam filter',
              parameters: null,
            },
          },
          {
            type: 'function',
            function: {
              name: 'deny_message' as const,
              description: 'Deny the message to get past the spam filter',
              parameters: null,
            },
          },
        ],
      },
      options: {
        maxErrorCorrectionTries: 2,
      },
    })

    const resultPromise = chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a message, and your task is to decide if that message is spam.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: 'function',
    })

    await expect(resultPromise).rejects.toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls.length).toEqual(3)
  })

  test('can answer a structured chat from the cache', async () => {
    mockCacheAdapterGet.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: null,
        tool_calls: [
          {
            type: 'function',
            id: 'call_2oKm3MB182UOdulkNKR5kpak',
            function: {
              name: 'return_sentiment',
              arguments: '{"sentiment":"positive"}',
            },
          },
        ],
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      cacheAdapter: mockCacheAdapter,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
      options: { timeout: 5 * 1000, maxRetries: 2 },
    })

    const result = await chat.send({
      messages: [
        {
          role: 'system',
          content:
            'You will be provided with a tweet, and your task is to classify its sentiment as positive, neutral, or negative.',
        },
        { role: 'user', content: 'I loved the new Batman movie!' },
      ],

      tool_choice: { type: 'function', function: { name: 'return_sentiment' } },
    })

    expect(result).toMatchSnapshot()
    expect(chat.getMessages()).toMatchSnapshot()
    expect(mockChatCompletionsCreate.mock.calls).toEqual([])

    expect(mockCacheAdapterGet.mock.calls).toMatchSnapshot()
    expect(mockCacheAdapterSet.mock.calls).toEqual([])
  })

  test('can generate a light JSON schema out of a Zod schema', async () => {
    const Schema = z.array(
      z.object({
        foo: z.string().describe('Foo'),
        bar: z.array(z.object({ baz: z.string().describe('Baz') })).describe('Bar'),
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: { model: 'gpt-4-turbo-preview' },
    })

    expect(chat._zodToJsonSchema(Schema)).toMatchSnapshot()
  })

  test('has typed tool_choice function names', async () => {
    mockChatCompletionsCreate.mockResolvedValueOnce(
      generateMockChatCompletionCreateResult({
        role: 'assistant',
        content: 'Mock assistant message',
      })
    )

    const chat = new StructuredChat({
      openai: mockOpenAi,
      params: {
        model: 'gpt-4-turbo-preview',
        tools: [
          {
            type: 'function',
            function: {
              name: 'return_sentiment' as const,
              description: 'Return the sentiment of the tweet',
              parameters: z.object({
                /** The sentiment of the tweet */
                sentiment: z
                  .enum(['positive', 'neutral', 'negative'])
                  .describe('The sentiment of the tweet'),
              }),
            },
          },
        ],
      },
    })

    const resultPromise = chat.send({
      messages: [{ role: 'system', content: 'Mock system message' }],

      tool_choice: {
        type: 'function',
        function: {
          // @ts-expect-error The function name does not exist
          name: 'foo_function',
        },
      },
    })

    await expect(resultPromise).rejects.toBeTruthy()
  })
})
